package pltag.german;

import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class HeadFinder {
	
	private HeadPercolTable pt;
	
	private int nodesAll;
	private int nodesPT;
	private int nodesDef;
	
	
	public HeadFinder(HeadPercolTable table) {
		pt = table;
		
		nodesAll = 0;
		nodesPT = 0;
		nodesDef = 0;
	}
	
	public void printStats() {
		System.out.println(stringStats());
	}
	
	public String stringStats() {
		nodesPT = pt.getCount();
		int accurate = nodesAll - nodesPT - nodesDef;
		StringBuffer sb = new StringBuffer();
		sb.append("\n### Head Identification Statistics ###\n");
		sb.append("All nodes considered: " + nodesAll + "\n");
		sb.append("Informed head child identification: " + accurate 
							+ "(" + 100*(double)accurate/nodesAll + "%)\n");
		sb.append("Percolation table head child identification: " + nodesPT 
				+ "(" + 100*(double)nodesPT/nodesAll + "%)\n");
		sb.append("Default head child identification: " + nodesDef 
				+ "(" + 100*(double)nodesDef/nodesAll + "%)\n");
		return sb.toString();
	}

	public void annotateHeads(TigerCorpus corpus) {
		for (TigerTree tree : corpus.getTrees().values()) {
			annotateHeads(tree.getRoot());
		}
		
	}

	public void annotateHeads(TreeNode node) {
		
		for (TreeNode child : node.getChildren()) {
			annotateHeads(child);
		}
		
		String phrasalCat = node.getPhrasalCategory();
		boolean found = false;
		nodesAll++;

		if (Utils.isCoord(node)) {
			found = cFindHead(node);
		}
		else if (phrasalCat.equals("VNODE")) {
			found = vnodeFindHead(node);
		}
		else if (phrasalCat.equals("AA")) {
			found = aaFindHead(node);
		}
		else if (phrasalCat.equals("AC")) {
			found = acFindHead(node);
		}
		else if (phrasalCat.equals("AP")) {
			found = apFindHead(node);
		}
		else if (phrasalCat.equals("AVP")) {
			found = avpFindHead(node);
		}
		else if (phrasalCat.equals("CH")) {
			found = chFindHead(node);
		}
		else if (phrasalCat.equals("DL")) {
			found = dlFindHead(node);
		}
		else if (phrasalCat.equals("DP")) {
			found = dpFindHead(node);
		}
		else if (phrasalCat.equals("ISU")) {
			found = isuFindHead(node);
		}
		else if (phrasalCat.equals("MTA")) {
			found = mtaFindHead(node);
		}
		else if (phrasalCat.equals("NM")) {
			found = nmFindHead(node);
		}
		else if (phrasalCat.equals("NP")) {
			found = npFindHead(node);
		}
		else if (phrasalCat.equals("PCC")) {
			found = pccFindHead(node);
		}
		else if (phrasalCat.equals("PN")) {
			found = pnFindHead(node);
		}
		else if (phrasalCat.equals("PP")) {
			found = ppFindHead(node);
		}
		else if (phrasalCat.equals("QL")) {
			found = qlFindHead(node);
		}
		else if (phrasalCat.equals("S")) {
			found = sFindHead(node);
		}
		else if (phrasalCat.equals("VP")) {
			found = sFindHead(node);
		}
		else if (phrasalCat.equals("VZ")) {
			found = vzFindHead(node);
		}
		
		//if (!found && phrasalCat.matches("VNODE")) {
		//	System.err.println(phrasalCat + ": no head child found");
		//	System.err.println(Utils.subtreeToString(node));
		//}
		
		// if no head was found (and this is not a leaf node),
		// we take the left-most child, which is not a trace, to be the head
		ArrayList<TreeNode> children = node.getChildren();
		if (!found && children.size() > 0) {
			for (TreeNode child : children) {
				if (child.isLeaf() && ((LeafNode)child).isTrace()) {
					continue;
				}
				node.setHeadChild(child);
				child.setHead();
				found = true;
				nodesDef++;
				break;
			}			
		}
		// if  still there was no head found, we just take the left-most child
		// (no matter whether it is a trace)
		if (!found && children.size() > 0) {
			node.setHeadChild(children.get(0));
			children.get(0).setHead();
			nodesDef++;
		}
			
	}
	
	private boolean vnodeFindHead(TreeNode node) {
		ArrayList<TreeNode> children = node.getChildren();
		// if there is only one child, it is the head
		if (children.size() == 1) {
			node.setHeadChild(children.get(0));
			children.get(0).setHead();
			return true;
		}
		// else: S, DL or CS constitute the head
		// if there is none, the first category from the left, which is not
		// punctuation, is the head
		int first = -1;
		for (int i = 0; i < children.size(); i++) {
			TreeNode child  = children.get(i);
			if (child.getPhrasalCategory().matches("S|DL|CS") ) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
			if (first == -1 && !child.getPhrasalCategory().matches("\\$,|\\$\\.|\\$\\*LRB\\*")) {
				first = i;
			}
		}
		if (first > -1) {
			node.setHeadChild(children.get(first));
			children.get(first).setHead();
			return true;
		}
		return false;
	}

	private boolean aaFindHead(TreeNode node) {
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.getFunctionalCategory().matches("HD")) {
				node.setHeadChild(child);
				child.setHead();
				found = true;
				break;
			}
		}
		if (!found) {
			found = pt.setHead(node);
		}
		return found;
	}
	
	private boolean acFindHead(TreeNode node) {
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.getFunctionalCategory().matches("HD")) {
				node.setHeadChild(child);
				child.setHead();
				found = true;
				break;
			}
		}
		//if (!found) {
		//	found = pt.setHead(node);
		//}
		return found;
	}

	private boolean apFindHead(TreeNode node) {
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.getFunctionalCategory().matches("HD|PH")) {
				node.setHeadChild(child);
				child.setHead();
				found = true;
				break;
			}
		}
		if (!found) {
			found = pt.setHead(node);
		}
		return found;		
	}
	
	private boolean avpFindHead(TreeNode node) {
		// we consider the left-most AVC to be the head.
		// optimally this will be a multi-anchored etree later
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (int i = 0; i <children.size(); i++) {
			TreeNode child = children.get(i);
			if (child.getFunctionalCategory().matches("HD|PH|AVC")) {
				node.setHeadChild(child);
				child.setHead();
				found = true;
				break;
			}
		}
		if (!found) {
			found = pt.setHead(node);
		}
		return found;		
	}
	
	private boolean chFindHead(TreeNode node) {
		// we consider the right-most UC to be the head.
		// optimally this will be a multi-anchored etree later
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (int i = children.size()-1; i > -1; i--) {
			TreeNode child = children.get(i);
			if (child.getFunctionalCategory().matches("UC")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		// not found yet
		found = pt.setHead(node);
		return found;
	}
	
	private boolean dlFindHead(TreeNode node) {
		// punctuation ($, or $. or $*LRB* : or $*LRB* -) is the head
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.getPhrasalCategory().matches("\\$\\.|\\$,")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
			if (child.getPhrasalCategory().equals("$*LRB*") &&
					((LeafNode)child).getLeafString().matches(":|-")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		// not found yet
		found = pt.setHead(node);
		return found;
	}
	
	private boolean dpFindHead(TreeNode node) {
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.getFunctionalCategory().matches("HD")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		// not found yet
		found = pt.setHead(node);
		return found;
	}
	
	private boolean isuFindHead(TreeNode node) {
		// we consider the left-most UC to be the head.
		// optimally this will be a multi-anchored etree later
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.getFunctionalCategory().matches("UC")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		// not found yet
		found = pt.setHead(node);
		return found;
	}
	
	private boolean mtaFindHead(TreeNode node) {
		// we consider the left-most ADC to be the head.
		// optimally this will be a multi-anchored etree later
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (int i = 0; i < children.size(); i++) {
			TreeNode child = children.get(i);
			if (child.getFunctionalCategory().matches("ADC")) {
				node.setHeadChild(child);
				child.setHead();
				found = true;
				break;
			}
		}
		if (!found) {
			found = pt.setHead(node);
		}
		return found;
	}
	
	private boolean nmFindHead(TreeNode node) {
		// we consider the right-most NN-NMC or, if there is no NN, 
		// the right-most NMC to be the head.
		// maybe this will be a multi-anchored etree later
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		// right-most NN-NMC
		for (int i = children.size()-1; i > -1; i--) {
			TreeNode child = children.get(i);
			if (child.isLeaf() && ((LeafNode)child).isTrace()) {
				continue;
			}
			if (child.getFunctionalCategory().matches("NMC") &&
					child.getPhrasalCategory().matches("NN")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		// right-most NMC
		for (int i = children.size()-1; i > -1; i--) {
			TreeNode child = children.get(i);
			if (child.isLeaf() && ((LeafNode)child).isTrace()) {
				continue;
			}
			if (child.getFunctionalCategory().matches("NMC")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		// not found yet
		found = pt.setHead(node);
		return found;
	}

	private boolean npFindHead(TreeNode node) {
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.isLeaf() && ((LeafNode)child).isTrace()) {
				continue;
			}
			if (child.getFunctionalCategory().matches("PH|NKHD|HD")) {
				node.setHeadChild(child);
				child.setHead();
				found = true;
				break;
			}
		}
		if (!found) {
			found = pt.setHead(node);
		}
		return found;
	}
	
	private boolean pccFindHead(TreeNode node) {
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.getFunctionalCategory().matches("CM")) {
				node.setHeadChild(child);
				child.setHead();
				found = true;
				break;
			}
		}
		return found;
	}
	
	private boolean pnFindHead(TreeNode node) {
		// if there are several PNCs,
		// we consider the right-most PNC to be the head.
		// optimally this will be a multi-anchored etree later
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (int i = children.size()-1; i > -1; i--) {
			TreeNode child = children.get(i);
			if (child.getFunctionalCategory().matches("PNC")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		// not found yet
		found = pt.setHead(node);
		return found;
	}
	
	private boolean ppFindHead(TreeNode node) {
		// if there are several ACs,
		// we consider the left-most AC to be the head.
		// optimally this will be a multi-anchored etree later
		boolean found = false;
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.isLeaf() && ((LeafNode)child).isTrace()) {
				continue;
			}
			if (child.getFunctionalCategory().matches("PH|AC")) {
				node.setHeadChild(child);
				child.setHead();
				found = true;
				break;
			}
		}
		if (!found) {
			found = pt.setHead(node);
		}
		return found;
	}
	
	private boolean qlFindHead(TreeNode node) {
		return pt.setHead(node);
	}
	
	private boolean sFindHead(TreeNode node) {
		boolean found = false;
		int hasVPOC = -1;
		int hasAux = -1;
		int someOtherHead = -1; // used for "exceptions" that do not really follow the "standard" annotation scheme
		ArrayList<TreeNode> children = node.getChildren();
		for (int i = 0; i < children.size(); i++) {
			TreeNode child = children.get(i);
			if (child.getFunctionalCategory().matches("HD") &&
					child.getPhrasalCategory().matches("(VA|VM).*")) {
				hasAux = i;
			}
			if (child.getFunctionalCategory().matches("OC") &&
					child.getPhrasalCategory().matches("VP|CVP") && !child.isFiller()) {
				hasVPOC = i;
			}
			if (child.getFunctionalCategory().matches("HD") &&
					child.getPhrasalCategory().matches("VV.*")) {				
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
			if (child.getFunctionalCategory().matches("HD") &&
					child.getPhrasalCategory().matches("VZ")) {
				TreeNode granChild = child.getHeadChild();
				if (granChild != null && granChild.getPhrasalCategory().matches("VV.*")) {
					node.setHeadChild(child);
					child.setHead();
					return true;
				}
				if (granChild != null && granChild.getPhrasalCategory().matches("(VA|VM).*")) {
					hasAux = i;
				}
			}
			if (child.getFunctionalCategory().matches("HD") &&
					child.getPhrasalCategory().matches("CVP")) {
				// we need to find out whether the "head" (i.e. the footnode) of the CVP is VA or VV
				for (TreeNode grandchild : child.getChildren()) {
					if (grandchild.getFunctionalCategory().equals("CJ")) {
						// this is probably going to be the footnode later.
						// so go down the head trunc.
						while (grandchild != null && !grandchild.isLeaf()) {
							grandchild = grandchild.getHeadChild();
						}
						if (grandchild != null && grandchild.getPhrasalCategory().matches("VV.*")) {
							node.setHeadChild(child);
							child.setHead();
							return true;
						}
						if (grandchild != null && grandchild.getPhrasalCategory().matches("(VA|VM).*")) {
							hasAux = i;
							break;
						}
					}
				}
			}
			if (child.getFunctionalCategory().matches("HD|SVP")) {
				someOtherHead = i;
			}
		}
		if (hasAux > -1 && hasVPOC > -1) {
			if (node.getPhrasalCategory().matches("S")) {
				TreeNode child = children.get(hasVPOC);
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
			if (node.getPhrasalCategory().matches("VP")) {
				TreeNode child = children.get(hasAux);
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		if (hasAux > -1) { //&& there is no VP-OC
			TreeNode child = children.get(hasAux);
			node.setHeadChild(child);
			child.setHead();
			return true;
		}
		// if were are still here, S does not have HD
		if (hasVPOC > -1) {
			TreeNode child = children.get(hasVPOC);
			node.setHeadChild(child);
			child.setHead();
			return true;
		}
		// if we are still here, the sentence might be a bit "weird", but we still
		// try to give this node a sensible head.
		if (someOtherHead > -1) {
			TreeNode child = children.get(someOtherHead);
			node.setHeadChild(child);
			child.setHead();
			return true;
		}
		
		// TODO head percolation table needs to be adapted for S
		//found = pt.setHead(node);

		return found;
	}
	
	private boolean vzFindHead(TreeNode node) {
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.getFunctionalCategory().matches("HD")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}

		return pt.setHead(node);
	}
	
	private boolean cFindHead(TreeNode node) {
		ArrayList<TreeNode> children = node.getChildren();
		for (TreeNode child : children) {
			if (child.isLeaf() && ((LeafNode)child).isTrace()) {
				continue;
			}
			if (child.getFunctionalCategory().matches("CD")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
		}
		// no coordinating conjunction
		// let's take the right-most comma as the head
		int otherPunct = -1;
		for (int i = children.size()-1; i > -1; i--) {
			TreeNode child  = children.get(i);
			if (child.getPhrasalCategory().matches("\\$,")) {
				node.setHeadChild(child);
				child.setHead();
				return true;
			}
			if (child.getPhrasalCategory().equals("$*LRB*") &&
					((LeafNode)child).getLeafString().matches("-|/|\\*LRB\\*|:")) {
				otherPunct = i;
			}
			else if (child.getPhrasalCategory().equals("$.") &&
					((LeafNode)child).getLeafString().matches("\\.|:")) {
				otherPunct = i;
			}
		}
		// no comma
		// let's take the other punctuation as the head
		if (otherPunct > -1) {
			TreeNode child = children.get(otherPunct);
			node.setHeadChild(child);
			child.setHead();
			return true;
		}

		return pt.setHead(node);
	}

}
