package pltag.german;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import pltag.german.regularadjunction.GrammarExtractorRegAdj;
import pltag.german.regularadjunction.HeadOutwardBinarizer;
import pltag.german.regularadjunction.PredictionTreeExtractorRegAdj;
import pltag.german.sisteradjunction.GrammarExtractorSisAdj;
import pltag.german.sisteradjunction.PredictionTreeExtractorSisAdj;


public class TAGCorpus {
	
	private static File lexiconFile; 
	private static File goldStandardFile; 
	private static File derivationFile;
	private static Writer lexWriter = null;
	private static Writer gsWriter = null;
	private static Writer derivationWriter = null;
	
	private static File traceGoodFile;
	private static File traceBadFile;
	private static Writer traceGoodWriter = null;
	private static Writer traceBadWriter = null;
	private int numTraceGood = 0;
	private int numTraceBad = 0;
	private int numTraceBadMod = 0;
	
	private static File predictionFile;
	private static Writer predictionWriter = null;
	
	private static File infoFile;
	private static Writer infoWriter = null;
	
	private static String sourcePath; 
	private static String outputPath;
	private static String outputAffix;
	private static String corpus = "";
	
	private static String inputTigerPennName = "tiger_release_aug07_ID_utf8_RB.penn";
	private static String inputTigerNegraName = "tiger_release_aug07.export";
	private static String inputTigerHeadPercTableName = "head-percol-table-tiger.txt";
	
	private boolean multiAnchoredTrees = false;
	private boolean mainAnchorAnywhereInMAT = false;
	private boolean motherLevelForSisterAdj = true;
	private int sisterAdjOnlyOnCompleteNodes = 3;
	private boolean regularAdjunction = false;
	private int deleteTraces = 0;
	
	
	public boolean debug = false;
	
	ArrayList<LexEntry> lexicon = new ArrayList<LexEntry>();
	
	
	public TAGCorpus(String[] args) {
		long start = System.currentTimeMillis();
		//Utils.printMemory("Start");
		
		try {
			parseCommandLineOptions(args);
		} catch (Exception e2) {
			e2.printStackTrace();
			System.exit(0);
		}
		
		outputPath = outputPath + File.separator + outputAffix;
		File outputPathDir = new File(outputPath);
		outputPathDir.mkdir();
		
		lexiconFile = new File(outputPath + File.separator + "Lexicon" + outputAffix + ".txt");
		goldStandardFile = new File(outputPath + File.separator + "GoldStandard" + outputAffix + ".txt");
		derivationFile = new File(outputPath + File.separator + "Derivation" + outputAffix + ".txt");
		try {lexWriter = new BufferedWriter(new FileWriter(lexiconFile));} catch (IOException e1) {e1.printStackTrace();}
		try {gsWriter = new BufferedWriter(new FileWriter(goldStandardFile));} catch (IOException e1) {e1.printStackTrace();}
		try {derivationWriter = new BufferedWriter(new FileWriter(derivationFile));} catch (IOException e1) {e1.printStackTrace();}
		
		traceGoodFile = new File(outputPath + File.separator + "TracesGood" + outputAffix + ".txt");
		traceBadFile = new File(outputPath + File.separator + "TracesUnbound" + outputAffix + ".txt");
		try {traceGoodWriter = new BufferedWriter(new FileWriter(traceGoodFile));} catch (IOException e1) {e1.printStackTrace();}
		try {traceBadWriter = new BufferedWriter(new FileWriter(traceBadFile));} catch (IOException e1) {e1.printStackTrace();}
		
		predictionFile = new File(outputPath + File.separator + "PredictionLexicon" + outputAffix + ".txt");
		try {predictionWriter = new BufferedWriter(new FileWriter(predictionFile));} catch (IOException e1) {e1.printStackTrace();}
		
		infoFile = new File(outputPath + File.separator + "info" + outputAffix + ".txt");
		try {infoWriter = new BufferedWriter(new FileWriter(infoFile));} catch (IOException e1) {e1.printStackTrace();}
		
		
		File tigerPennFile = new File(sourcePath + File.separator + inputTigerPennName + corpus);
		File tigerNegraFile = new File(sourcePath + File.separator + inputTigerNegraName);
		File tigerHeadPercolFile = new File(sourcePath + File.separator + inputTigerHeadPercTableName);
		
		//Utils.printMemory("Before head percol table");
		
		// Read in the head percolation table (Rehbein, 2009);
		// instantiate the head finder which makes use amongst other of the
		// head percolation table
		HeadPercolTable hpt = new HeadPercolTable(tigerHeadPercolFile);
		HeadFinder headFinder = new HeadFinder(hpt);
		
		//Utils.printMemory("Before tiger");
		
		// Read the treebank in Penn format into a tree structure;
		// enrich it with information from the same treebank in the Negra format:
		// * lemma
		// * morphology
		// * secondary edges
		TigerCorpus tiger = new TigerCorpus(tigerPennFile, tigerNegraFile);
		
		//Utils.printMemory("After tiger");
		
		TigerTree tree;
		// treat the sentences in original order (more convenient for printing)
		List<Integer> tigerTreeNumbers = new ArrayList<Integer>(tiger.getTrees().keySet());
		Collections.sort(tigerTreeNumbers);
		
		// Sister-adjunction OR regular adjunction 
		AbstractGrammarExtractor extractor;
		AbstractPredictionTreeExtractor predTreeExtractor;
		if (!regularAdjunction) {
			extractor = new GrammarExtractorSisAdj(multiAnchoredTrees);
			predTreeExtractor = new PredictionTreeExtractorSisAdj(multiAnchoredTrees, 
					sisterAdjOnlyOnCompleteNodes);	
		}
		else { //if (regularAdjunction) {
			extractor = new GrammarExtractorRegAdj(multiAnchoredTrees);
			predTreeExtractor = new PredictionTreeExtractorRegAdj(multiAnchoredTrees);
			regularAdjunction = true;
		}
		
		// Initializations
		ArrayList<LexEntry> currLexicon;
		ArrayList<ElementaryTree> currElementTree = new ArrayList<ElementaryTree>();
		ArrayList<PredictionTree> currPredLexicon = new ArrayList<PredictionTree>();
		
		MorphAnnotator morphAnnotator = new MorphAnnotator();
		DerivationFinder derivFinder = new DerivationFinder();
		
		int sent = 0;
		int problemSent = 0;
		int sentencesWithUnboundTraces = 0;
		int sentencesWithSecondaryEdge = 0;
		int treeNodesBeforeConversionSum = 0;
		int treeNodesAfterConversionSum = 0;
		
		ElementaryTree r = null;
		ElementaryTree eTree = null;
		CompositeTree partialTree = null;
		StringBuffer errorBuffer = null;
		TigerTree treeCopy = null;
		LTAGDerivation deriv = null;
		
		System.out.println("Starting treebank transformation and lexicon extraction ...");
		
		THIS_TREE: for (Integer s : tigerTreeNumbers) {
			if (s % 1000 == 0) {
				System.out.println(s);
				//Utils.printMemory("Ex");
				if (debug) {Utils.printMemory("In Conversion"); }
			}
			//System.out.println(s);
			//if (s != 659) continue;
			sent++;
			tree = tiger.getTrees().get(s);  
			treeNodesBeforeConversionSum += tree.countNodes();
			
			if (tree.getSecondaryEdges().size() > 0) {
				sentencesWithSecondaryEdge++;
			}
			
			treeCopy = new TigerTree(tree.getSourceString());
		
			// PP -> P NP
			tree.insertNPUnderPP(tree.getRoot());
			// Introduce maximal projections: AVP, AP, VP, DP
			tree.introduceMaxProjections(tree.getRoot());
			// Disambiguate NPs and introduces NP projections
			tree.introduceNPs(tree.getRoot());
			// Introduce CCP
			tree.introducePCC(tree.getRoot());
			// Re-attach some fillers
			tree.reattachSomeFillers();
			
			// Find the head child of each node.
			// Do this after conversion, then there will be less rules
			headFinder.annotateHeads(tree.getRoot());
			
			// some problems with coordinating structures are fixed here
			tree.fixCoordinations(tree.getRoot());
			
			// argument-modifier distinction
			tree.determineSiblingStatus();
			
			// identify predicative auxiliary trees and
			// insert some S levels to bind traces
			tree.identifyAuxiliaryTrees(tree.getRoot());
			
			// Do sth about punctuation
			tree.reattachPunctuation();
			// Remove VNODE if possible
			tree.removeRoot();
			
			// Delete traces
			if (deleteTraces == 1 || deleteTraces == 3) {
				tree.deleteModifierTraces(treeCopy);
			}
			else if (deleteTraces == 2) {
				tree.deleteAllTraces(treeCopy);
			}
			
		
			if (regularAdjunction) {
				// head-outward binarization
				HeadOutwardBinarizer binarizer = new HeadOutwardBinarizer();
				binarizer.binarize(tree);
			}
			
			ArrayList<String> tracesToDelete = new ArrayList<String>();
			boolean finished = false;
	
			while (!finished) {
				
				if (tracesToDelete.size() > 0) {
					tree.deleteTraces(tracesToDelete, treeCopy, regularAdjunction);
				}
				
				// extract canonical elementary trees:
				// * create "lexicon entries" (pointers to the nodes that belong to etrees)
				currLexicon = extractor.createLexicon(tree);
				currElementTree = new ArrayList<ElementaryTree>();
	
				// * turn the lexicon entries into proper elementary tree structures
				for (LexEntry lex : currLexicon) {
					eTree = new ElementaryTree(lex);
					currElementTree.add(eTree);
					// remember the eTree which provides the root
					if (eTree.getRoot().getID() == tree.getRoot().getID() && 
							eTree.getType() == NodeStatus.ARG) { // ??
						r = eTree;
					}				
				}
				
				// morphology
				morphAnnotator.setMorphology(tree);
				for (ElementaryTree e : currElementTree) {
					morphAnnotator.setMorphologyFromTree(e.getRoot());
				}
				
				
				// for multi-anchored trees (MAT)
				if (multiAnchoredTrees) {
					// first fix the lexicon entries (for deeper embedded trees that should go into one)
					fixLexEntriesForMAT(currLexicon);
					// put the elementary trees together to multi-anchored trees (MAT)
					for (ElementaryTree e : currElementTree) {
						boolean success = e.turnIntoMultiAnchoredTree(mainAnchorAnywhereInMAT);
						if (!success) {
							// it is not a real error here. elementary trees were not combined
							// because the main anchor would have not been the leftmost one in the eTree.
						}
					}
				}
				
				// Check whether the elementary trees can be put together again
				// to form the original Tiger tree;
				// we start with the eTree which provides the root
				partialTree = new CompositeTree(r);
				errorBuffer = new StringBuffer();
				for (ElementaryTree eTree2 : currElementTree) {
					if (eTree2 != r && !eTree2.isPartOfMAT()) {
						if (!Utils.isWellFormed(eTree2, errorBuffer, tree.getTraceList())) break;
						
						partialTree.tryToIntegrate(eTree2, errorBuffer);
					}
				}
				if (errorBuffer.length() == 0) { // no problems encountered so far, go on
					boolean success = partialTree.tryMore(errorBuffer);
					if (!success) {
						errorBuffer.append("### Error in putting together tree.");
					}
				}
				boolean match = false;
				if (errorBuffer.length() == 0) { // still no problems encountered so far
					// check whether the rebuilt tree corresponds to the Tiger tree
					match = partialTree.correspondsTo(tree, errorBuffer);
				}
				if (errorBuffer.length() != 0 || !match) { // some problem occurred
					problemSent++;
					System.err.println("### %% " + s + ":");
					System.err.println(errorBuffer.toString());
					if (debug) System.err.println(partialTree.toString()); 
					continue THIS_TREE;
				}
				
				//if (deleteTraces == 3) {
				//	deleteTraceFillerNonPairs(tree, partialTree, treeCopy, currElementTree);
				//}
				
				// derivation
				deriv = derivFinder.computeDerivation(partialTree, currElementTree);
				
				
				// extract prediction trees
				StringBuffer predErrorString = new StringBuffer();
				currPredLexicon = predTreeExtractor.createPredLexicon(tree, predErrorString);
				if (predErrorString.length() != 0) {
					problemSent++;
					predErrorString.insert(0, "### Error in prediction trees : ");
					System.err.println("### %% " + s + ":");
					System.err.println(predErrorString.toString()); 
					continue;
				}
				
				if (deleteTraces == 3) {
					tracesToDelete = checkTraceFillerNonPairs(tree, currElementTree);
					if (tracesToDelete.size() == 0 ) {
						finished = true;
					}
					else {
						// try again
						r = null;
						eTree = null;
						partialTree = null;
						errorBuffer = null;
						currElementTree.clear();
						currPredLexicon.clear();
						deriv = null;
					}
				}
				else {
					finished = true;
				}
			}
			
			// have a look at the traces
			try {
				boolean sentenceConcerned = checkTraces(tree.getTraceList(), currElementTree, errorBuffer);
				if (errorBuffer.length() != 0) {
					problemSent++;
					System.err.println("### %% " + s + ":");
					System.err.println(errorBuffer.toString()); 
				}
				if (sentenceConcerned) {
					sentencesWithUnboundTraces++;
				}
			}
			catch(Exception e) {
				System.err.println("### Print traces" + s);
				e.printStackTrace();
			}
			
			// print gold standard
			try {
				printGoldStandard(s, partialTree, tree, treeCopy);
			}
			catch(Exception e) {
				System.err.println("### PrintGS" + s);
				e.printStackTrace();
			}
			
			// PRINT sentence ID
			try {
				lexWriter.append("%%% " + s + "\n");
			} 
			catch (IOException e) {e.printStackTrace();}
			
			// print prediction lexicon
			for (PredictionTree pred : currPredLexicon) {
				
				try {
					printLex(pred.toStringLex(motherLevelForSisterAdj, tree.getTraceList()));
					printPredLexSeparately(pred, tree.getTraceList());
					
				}
				catch(Exception e) {
					System.err.println("### PrintPredLex " + s);
					e.printStackTrace();
				}
			}
			
			// print lexicon
			for (ElementaryTree eTree2 : currElementTree) {
				try {
					printLex(eTree2.toStringLex(motherLevelForSisterAdj, tree.getTraceList()));
				}
				catch(Exception e) {
					System.err.println("### PrintLex " + s);
					e.printStackTrace();
				}
			}
			
			// print derivation
			try {
				derivationWriter.append("%%% " + s + "\n");
				derivationWriter.append(deriv.toStringForOutput());
			} 
			catch (IOException e) {e.printStackTrace();}
			
			treeNodesAfterConversionSum += tree.countNodes();
			
			r = null;
			eTree = null;
			partialTree = null;
			errorBuffer = null;
			currElementTree = null;
			currPredLexicon = null;
			
			// we won't need this sentence anymore, so let's free some memory
			tiger.getTrees().remove(s);
		}
		
		try {lexWriter.close();} catch (IOException e) {e.printStackTrace();}
		try {gsWriter.close();} catch (IOException e) {e.printStackTrace();}
		try {derivationWriter.close();} catch (IOException e) {e.printStackTrace();}
		try {traceGoodWriter.close();} catch (IOException e) {e.printStackTrace();}
		try {traceBadWriter.close();} catch (IOException e) {e.printStackTrace();}
		try {predictionWriter.close();} catch (IOException e) {e.printStackTrace();}
			
		
		long end = System.currentTimeMillis();
		int nodesDiff = treeNodesAfterConversionSum-treeNodesBeforeConversionSum;
		
		try {
			writeProgramInfo();
		
			System.out.println("Execution time was "+(end-start)+" ms.");
			
			writeToInfoAndOut("\nNumber of sentences: " + sent);
			writeToInfoAndOut("Number of sentences with secondary edges: " + sentencesWithSecondaryEdge + 
						" (" + 100*(double)sentencesWithSecondaryEdge/sent + "%)");
			writeToInfoAndOut("Number of error sentences: " + problemSent + "\n");
			writeToInfoAndOut("Number of bound traces: " + numTraceGood);
			writeToInfoAndOut("Number of unbound traces: " + numTraceBad);
			writeToInfoAndOut("\t\tthereof non-modifiers: " + (numTraceBad - numTraceBadMod));
			writeToInfoAndOut("Number of sentences with unbound argument trace(s): " + sentencesWithUnboundTraces +
						" (" + 100*(double)sentencesWithUnboundTraces/sent + "%)\n");
			if (deleteTraces == 1) { // deleteModTraces
				writeToInfoAndOut("Plus many unbound modifier traces that do not occur in this " +
						"statistics because they were deleted.\n\n");
			}
			if (deleteTraces == 2) { // deleteAllTraces
				writeToInfoAndOut("All traces were deleted!!!\n\n");
			}
			writeToInfoAndOut("Number of new nodes in Tiger trees: " + nodesDiff + 
					" (" + (100*(double)nodesDiff/treeNodesBeforeConversionSum) + "%)");
			writeToInfoAndOut("Number of new nodes in one tree, on average: " + (double)nodesDiff/(sent-problemSent));
			
			writeToInfoAndOut(headFinder.stringStats());
			
			infoWriter.close();
		}
		catch (IOException e) {
			e.printStackTrace();
		}
	}


	/************************************************************************/
	
	private void writeToInfoAndOut(String s) throws IOException {
		System.out.println(s);
		infoWriter.write(s + "\n");
	}
	
	private void writeProgramInfo() throws IOException {
		infoWriter.write("#### TAGCorpus.java ####\n\n");
		infoWriter.write("source path: " + sourcePath + "\n");
		infoWriter.write("output path: " + outputPath + "\n");
		infoWriter.write("output affix: " + outputAffix + "\n");
		infoWriter.write("sub-corpus: " + corpus + "\n\n");
		
		// prediction strategy (in case sister-adjunction is used)
		infoWriter.write("prediction strategy: ");
		if (sisterAdjOnlyOnCompleteNodes == 1) {
			infoWriter.write("a\n");
		}
		else if (sisterAdjOnlyOnCompleteNodes == 3) {
			infoWriter.write("b\n");
		}
		else if (sisterAdjOnlyOnCompleteNodes == 2) {
			infoWriter.write("c\n");
		}
		
		// MAT strategy
		infoWriter.write("MAT strategy: ");
		if (!multiAnchoredTrees && !mainAnchorAnywhereInMAT) {
			infoWriter.write("i\n");
		}
		else if (multiAnchoredTrees && !mainAnchorAnywhereInMAT) {
			infoWriter.write("ii\n");
		}
		else if (multiAnchoredTrees && mainAnchorAnywhereInMAT) {
			infoWriter.write("iii\n");
		}
		
		// adjunction
		infoWriter.write("adjunction (task): ");
		if (regularAdjunction) {
			infoWriter.write("regAdj\n");
		}
		else {
			infoWriter.write("sisAdj");
			if (!motherLevelForSisterAdj) {
				infoWriter.write(" chiangsSisAdj");
			}
			infoWriter.write("\n");
		}
		
		// traces
		infoWriter.write("trace deletion: ");
		if (deleteTraces == 0) {
			infoWriter.write("none\n");
		}
		else if (deleteTraces == 1) {
			infoWriter.write("deleteModTraces\n");
		}
		else if (deleteTraces == 2) {
			infoWriter.write("deleteAllTraces\n");
		}
		else if (deleteTraces == 3) {
			infoWriter.write("deleteTraceFillerNonPairs\n");
		}
		
		infoWriter.write("\n#############################################\n\n");
		
	}
	
	private void parseCommandLineOptions(String[] args) throws Exception {
		if (args.length < 3) {
			throw new Exception("Not enough arguments");
		}
		sourcePath = args[0];
		outputPath = args[1];
		outputAffix = args[2];
		String predStrat = "";
		String matStrat = "";
		String ourSisAdj = "";
		String task = "";
		for (int i = 3; i < args.length; i++) {
			if (args[i].equals("-corpus")) {
				if (i+1 < args.length && args[i+1].matches("train|dev|test")) {
					corpus = "." + args[i+1];
				}
				else {
					throw new Exception("No valid argument of \"-corpus\"");
				}
			}
			if (args[i].equals("-predStrat")) {
				if (i+1 < args.length && args[i+1].matches("a|b|c")) {
					predStrat = args[i+1];
				}
				else {
					throw new Exception("No valid argument of \"-predStrat\"");
				}
			}
			if (args[i].equals("-matStrat")) {
				if (i+1 < args.length && args[i+1].matches("i|ii|iii")) {
					matStrat = args[i+1];
				}
				else {
					throw new Exception("No valid argument of \"-matStrat\"");
				}
			}
			if (args[i].equals("-chiangsSisAdj")) {
				motherLevelForSisterAdj = false;
			}
			if (args[i].equals("-task")) {
				if (i+1 < args.length && args[i+1].matches("sisAdj|regAdj")) {
					task = args[i+1];
				}
				else {
					throw new Exception("No valid argument of \"-ourSisAdj\"");
				}
			}
			// delete traces?
			if (args[i].equals("-deleteModTraces")) {
				deleteTraces = 1;
			}
			if (args[i].equals("-deleteAllTraces")) {
				deleteTraces = 2;
			}
			if (args[i].equals("-deleteTraceFillerNonPairs")) {
				deleteTraces = 3;
			}
		}
		// prediction strategy (in case sister-adjunction is used)
		if (predStrat.equals("a")) {
			sisterAdjOnlyOnCompleteNodes = 1;
		}
		else if (predStrat.equals("b")) {
			sisterAdjOnlyOnCompleteNodes = 3;
		}
		else if (predStrat.equals("c")) {
			sisterAdjOnlyOnCompleteNodes = 2;
		}
		// MAT strategy
		if (matStrat.equals("i")) {
			multiAnchoredTrees = false;
			mainAnchorAnywhereInMAT = false;
		}
		else if (matStrat.equals("ii")) {
			multiAnchoredTrees = true;
			mainAnchorAnywhereInMAT = false;
		}
		else if (matStrat.equals("iii")) {
			multiAnchoredTrees = true;
			mainAnchorAnywhereInMAT = true;
		}
		// sister-adjunction or regular adjunction?
		if (task.equals("sisAdj")) {
			regularAdjunction = false;
		}
		else if (task.equals("regAdj")) {
			regularAdjunction = true;
		}
	}

	private void fixLexEntriesForMAT(ArrayList<LexEntry> currLexicon) {
		boolean change = true;
		while (change) {
			change = false;
			for (LexEntry lex : currLexicon) {
				if (lex.getAssociates().size() > 0) {
					for (LexEntry cand : currLexicon) {
						if (cand.getAssociates().contains(lex)) {
							cand.getAssociates().addAll(lex.getAssociates());
							lex.getAssociates().clear();
							change = true;
							break;
						}
					}
				}
			}
		}
	}
	
	private ArrayList<String> checkTraceFillerNonPairs(TigerTree tree,
			ArrayList<ElementaryTree> currLex) {
		
		ArrayList<String> tracesToDelete = new ArrayList<String>();
		
		// go through all the traces that occur in the tree
		Iterator<Map.Entry<String,Trace>> iter = tree.getTraceList().entrySet().iterator();
		while (iter.hasNext()) {
		    Map.Entry<String,Trace> entry = iter.next();
		    String traceID = entry.getKey();

			ElementaryTree traceTree = null;
			ElementaryTree fillerTree = null;
			
			// check in which eTree trace and filler occur.
			for (ElementaryTree eTree : currLex) {
				// don't count the traces in subtrees of MAT; they will be counted within the MAT
				if (eTree.isPartOfMAT()) {
					continue;
				}
				if (eTree.getTraceSet().containsKey(traceID)) {
					ElementaryTreeNode[] trace = eTree.getTraceSet().get(traceID);
					if (traceTree == null && trace[0] != null) {
						traceTree = eTree;
					}
					if (fillerTree == null && trace[1] != null) {
						fillerTree = eTree;
					}
					if (traceTree != null && fillerTree != null) {
						break;
					}
				}	
			}
			
			if (traceTree != null && fillerTree != null) {
				// if trace and filler occur in the same eTree, the trace is bound.
				// we are happy :) otherwise:
				if (traceTree != fillerTree) {
					tracesToDelete.add(traceID);
				}
			}
		}
		return tracesToDelete;
	}
	

	private boolean checkTraces(HashMap<String, Trace> traceList, 
					ArrayList<ElementaryTree> currLex, StringBuffer errorBuffer) throws IOException{
		
		boolean sentenceConcerned = false;
		
		// go through all the traces that occur in the tree
		for (String traceID : traceList.keySet()) {
			ElementaryTree traceTree = null;
			ElementaryTree fillerTree = null;
			// check in which eTree trace and filler occur.
			for (ElementaryTree eTree : currLex) {
				// don't count the traces in subtrees of MAT; they will be counted within the MAT
				if (eTree.isPartOfMAT()) {
					continue;
				}
				if (eTree.getTraceSet().containsKey(traceID)) {
					ElementaryTreeNode[] trace = eTree.getTraceSet().get(traceID);
					if (traceTree == null && trace[0] != null) {
						traceTree = eTree;
					}
					if (fillerTree == null && trace[1] != null) {
						fillerTree = eTree;
					}
					if (traceTree != null && fillerTree != null) {
						break;
					}
				}
			}
			if (traceTree != null && fillerTree != null) {
				// if trace and filler occur in the same eTree, the trace is bound.
				// we are happy :)
				if (traceTree == fillerTree) {
					numTraceGood++;
					traceGoodWriter.write(traceTree.toStringLex(traceList) + "\n");
				}
				else {
					numTraceBad++;
					// for the unbound traces, we do not care very much about the modifiers,
					// so we count them extra
					//if (traceTree.getRoot().getFunctionalCategory().matches(Utils.modifierFunctionalCats)) {
					if (traceList.get(traceID).getHole().getFunctionalCategory().matches(Utils.modifierFunctionalCats)) {
						numTraceBadMod++;
					}
					else if (!regularAdjunction
							&& traceTree.getType().equals(NodeStatus.MOD) && traceTree.getRoot().isTrace()) {
						// if the traceTree is only the trace itself and if it is a modifier tree,
						// we also consider it a modifier and count it extra
						numTraceBadMod++;
					}
					else if (regularAdjunction
							&& traceTree.getType().equals(NodeStatus.ADJ)
							&& traceTree.getAnchor().isTrace()) {
						numTraceBadMod++;
					}
					else {
						traceBadWriter.write(traceTree.toStringLex(traceList) + "\n");
						traceBadWriter.write(fillerTree.toStringLex(traceList) + "\n\n");
						sentenceConcerned = true;
					}
				}
				
			}
			else if (traceTree == null && fillerTree != null
					&& deleteTraces == 1
					&& traceList.get(traceID).getHole().getNodeStatus() == NodeStatus.MOD) {
				// modifier trace
			}
			else {
				errorBuffer.append("### Error in trace " + traceList.get(traceID));
			}
		}
		
		return sentenceConcerned;
		
	}

	private void printLex(String lexEntry) throws IOException{
		lexWriter.append(lexEntry + "\n");
	}
	
	private void printPredLexSeparately(PredictionTree pred, 
			HashMap<String, Trace> traceListTigerTree) throws IOException{
		for (ElementaryTree p : pred.getIndividualPredictionTrees()) {
			predictionWriter.append(p.toStringLex(motherLevelForSisterAdj, traceListTigerTree) + "\n");
		}
	}
	
	private void printGoldStandard(Integer i, CompositeTree tree, TigerTree tigerTree,
			TigerTree tigerTreeOriginal) throws IOException{
		StringBuffer sbLex = new StringBuffer();
		StringBuffer sbPos = new StringBuffer();
		StringBuffer sbTree = new StringBuffer();
		StringBuffer sbTreeOriginal = new StringBuffer();
		StringBuffer sbLexControl = new StringBuffer();
		
		printGoldStandardHelperTree(tree.getRoot(), sbTree, tigerTree.getSecondaryEdges(),
				tigerTree.getTraceList());
		printGoldStandardHelperTreeOriginal(tigerTreeOriginal.getRoot(), sbTreeOriginal, 
				tigerTreeOriginal.getSecondaryEdges(),tigerTreeOriginal.getTraceList());
		printGoldStandardHelperPos(tree.getRoot(), sbLex, sbPos);
		
		// Just some sanity check
		for (LeafNode leaf : tigerTree.getLeafNodes()) {
			if (leaf.isTrace()) {
				// CHANGE
				//sbLexControl.append(leaf.getPhrasalCategory());
				sbLexControl.append("*T*");
			}
			else {
				sbLexControl.append(leaf.getLeafString());
			}
			sbLexControl.append(" ");
		}
		if (!sbLexControl.toString().trim().equals(sbLex.toString().trim())) {
			System.err.println("### GS leaf nodes wrong " + i);
		}	
		
		gsWriter.append("%%% " + i + "\n");
		gsWriter.append(sbLex.toString().trim() + "\n");
		gsWriter.append(sbPos.toString() + "\n");
		gsWriter.append(sbTree.toString().trim() + "\n");
		gsWriter.append(sbTreeOriginal.toString().trim() + "\n");
	}
	
	private void printGoldStandardHelperPos(ElementaryTreeNode node, StringBuffer sbLex,
			StringBuffer sbPos) {
		
		if (node.isLeaf()) {
			String w = node.getLeaf();
			if (!node.isTrace()) {
				sbLex.append(w);
				sbPos.append(node.getPhrasalCategory());
				sbPos.append(" ");
				sbPos.append(w);
				sbPos.append(" ");
				sbPos.append(node.getLemma());
				sbPos.append("\t");
			}
			else {
				// CHANGE
				//sbLex.append(node.getPhrasalCategory());
				sbLex.append("*T*");
			}
			sbLex.append(" ");
		}
		else {
			for (ElementaryTreeNode child : node.getChildren()) {
				printGoldStandardHelperPos(child, sbLex, sbPos);
			}
		}
	}
	
	private void printGoldStandardHelperTree(ElementaryTreeNode node, 
			StringBuffer sbTree, ArrayList<SecondaryEdge> secEdges, HashMap<String, Trace> traceList) {

		sbTree.append("(");
		if (node.isTrace()) {   // CHANGE
			//sbTree.append("*T*");
			sbTree.append(node.getPhrasalCategory());
		}
		else {
			sbTree.append(node.getPhrasalCategory());
			sbTree.append(node.getMorphology());
			
			// filler
			// could also use getTreeNodeBottom
			if (node.getTreeNodeTop().isFiller() && traceList.containsKey(node.getTreeNodeTop().traceID)) {  // CHANGE
				String ind = node.getTreeNodeTop().traceID.replaceAll("\\*", "");
				sbTree.append("{").append(ind).append("}");
			}
		}
		
		// secondary edges	
		if (secEdges.size() > 0) {
			for (SecondaryEdge sec : node.getTreeNodeTop().getOutgoingSecondaryEdges()) {
				int i = secEdges.indexOf(sec);
				if (i >= 0) {
					sbTree.append("$$").append(i).append("-").append(sec.getLabel()).append("$$");
				}
			}
			for (SecondaryEdge sec : node.getTreeNodeBottom().getIncomingSecondaryEdges()) {
				int i = secEdges.indexOf(sec);
				if (i >= 0) {
					sbTree.append("$$").append(i).append("$$");
				}
			}
		}
		
		sbTree.append(" ");
		
		if (node.isLeaf()) {
			String w = node.getLeaf();
			if (!node.isTrace()) {				
				sbTree.append(w);
			}
			else {
				sbTree.append("-");
			}
		}
		else {
			for (ElementaryTreeNode child : node.getChildren()) {
				printGoldStandardHelperTree(child, sbTree, secEdges, traceList);
			}
		}
		
		sbTree.append(")");
	}
	
	private void printGoldStandardHelperTreeOriginal(TreeNode node, 
			StringBuffer sbTree, ArrayList<SecondaryEdge> secEdges, HashMap<String, Trace> traceList) {

		sbTree.append("(");
		if (node.isLeaf() && ((LeafNode)node).isTrace()) {   // CHANGE
			//sbTree.append("*T*");
			sbTree.append(node.getPhrasalCategory());
		}
		else {
			sbTree.append(node.getPhrasalCategory());
			sbTree.append(node.getMorphology());
			
			// filler
			if (node.isFiller() && traceList.containsKey(node.traceID)) {  // CHANGE
				String ind = node.traceID.replaceAll("\\*", "");
				sbTree.append("{").append(ind).append("}");
			}
		}
		
		// secondary edges	
		if (secEdges.size() > 0) {
			for (SecondaryEdge sec : node.getOutgoingSecondaryEdges()) {
				int i = secEdges.indexOf(sec);
				if (i >= 0) {
					sbTree.append("$$").append(i).append("-").append(sec.getLabel()).append("$$");
				}
			}
			for (SecondaryEdge sec : node.getIncomingSecondaryEdges()) {
				int i = secEdges.indexOf(sec);
				if (i >= 0) {
					sbTree.append("$$").append(i).append("$$");
				}
			}
		}
		
		sbTree.append(" ");
		
		if (node.isLeaf()) {
			String w = ((LeafNode)node).getLeafString();
			if (!((LeafNode)node).isTrace()) {				
				sbTree.append(w);
			}
			else {
				sbTree.append("-");
			}
		}
		else {
			for (TreeNode child : node.getChildren()) {
				printGoldStandardHelperTreeOriginal(child, sbTree, secEdges, traceList);
			}
		}
		
		sbTree.append(")");
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		TAGCorpus tag = new TAGCorpus(args);
	}

}
