package pltag.german.regularadjunction;


import pltag.german.AbstractGrammarExtractor;
import pltag.german.LeafNode;
import pltag.german.LexEntry;
import pltag.german.NodeStatus;
import pltag.german.TreeNode;
import pltag.german.Utils;

public class GrammarExtractorRegAdj extends AbstractGrammarExtractor {

	public GrammarExtractorRegAdj() {
		super();
	}

	public GrammarExtractorRegAdj(boolean multiAnchoredTrees) {
		super(multiAnchoredTrees);
	}

	@Override
	protected void extract(TreeNode node, LexEntry lex) {
		if (!currTigerTree.isBinary()) {
			System.err.println("### Tiger tree not binary for regular adjunction. Skipped.");
			return;
		}
		
		LeafNode leaf = extractSubtree(node, lex);
		lex.setLeafLink(leaf);
		currLexicon.add(lex);
	}
	
	private TreeNode extractPredAdj(TreeNode node, LexEntry lex) {
		lex.setAdjTree();
		LeafNode leaf = extractSubtree(node, lex, false);
		lex.setLeafLink(leaf);
		currLexicon.add(lex);
		return lex.getFoot();
	}
	
	private void addToLexEntry(TreeNode node, LexEntry lex) {
		extractSubtree(node, lex);
		// here we only add to the lex entry, but we do not put it
		// to the lexicon
	}
	
	private LeafNode extractSubtree(TreeNode node, LexEntry lex) {
		return extractSubtree(node, lex, true);
	}
	
	// the boolean is needed such that extractPredAdj does not go into an
	// endless loop
	private LeafNode extractSubtree(TreeNode node, LexEntry lex, boolean first) {

		// find the head trunk
		while (true) {

			// treat coordination and predicative auxiliary trees
			while (Utils.isCoord(node)
					|| (node.getNodeStatusBottom() == NodeStatus.PREDADJ && first)) {

				TreeNode foot;
				if (node.getNodeStatusBottom() == NodeStatus.PREDADJ) {
					foot = extractPredAdj(node, new LexEntry(node,
							multiAnchoredTrees));
				} 
				else {
					foot = extractCoordination(node);
				}
				if (foot == null) {
					break;
				}
				node = foot;
			}

			if (!node.isLeaf()) {
				TreeNode head = node.getHeadChild();
				
				if (node.getChildren().size() == 1) {
					head.setTop(lex);
					lex.addNode(head);
					node.setBottom(lex);
					lex.addNode(node);
					node = head;
				}
				else {
					// several children, check which configuration it is
					for (TreeNode child : node.getChildren()) {
						if (child.isHead()) continue;
						NodeStatus ns = child.getNodeStatus();
						if (ns == NodeStatus.ARG || ns == NodeStatus.MAARG) {
							if (child.isLeaf() && ((LeafNode)child).isTrace()) {
								// bind argument trace into the current lex entry
								addToLexEntry(child, lex);
							}
							else if (lex.isAdjTree() 
									&& currTigerTree.getFoot2Root().containsKey(child.getID())
									&& currTigerTree.getFoot2Root().get(child.getID()) == lex.getRoot()) {
								child.setTop(lex);
								lex.addNode(child);
								child.isFoot();
								lex.setFoot(child);
							}
							else {
								child.setTop(lex);
								lex.addNode(child);
								LexEntry newLex = new LexEntry(child, multiAnchoredTrees);
								if (ns == NodeStatus.MAARG) {
									lex.addAssociate(newLex);
								}
								extract(child, newLex);
							}
							head.setTop(lex);
							lex.addNode(head);
							node.setBottom(lex);
							lex.addNode(node);
							node = head;
						}
						else if (ns == NodeStatus.MOD) {
							TreeNode foot = extractAdj(child, new LexEntry(node, multiAnchoredTrees));
							node = foot;
						}
					}
				}
			} 
			else { // anchor found
				node.setBottom(lex);
				lex.addNode(node);
				return (LeafNode) node;
			}
		}
	}

	private TreeNode extractCoordination(TreeNode node) {
		LexEntry lex = new LexEntry(node, multiAnchoredTrees);
		lex.setAdjTree();
		//node.setBottom(lex);
		//lex.addNode(node);		
		
		// this is for ugly coordinated phrases that do not have a "real" head/anchor;
		// we do not finish this current coordination lex entry, but give
		// the node back to the "normal" extraction procedure.
		TreeNode headPath = node.getHeadChild();
		while (!headPath.isLeaf()) {
			if (!headPath.insertedForBinarization()) {
				return null;
			}
			headPath = headPath.getHeadChild();
		}
		//String headCat = headPath.getHeadChild().getPhrasalCategory();
		//if (!headCat.matches("KON")
		//		&& !Utils.isPunct(headCat)) {
		//	return null;
		//}
		
		// there will be no appropriate foot note in the coordination tree,
		// so we give the node back to the "normal" extraction procedure.
		if (!footAvailable(node)) {
			return null;
		}
		
		// special procedure for entweder-oder constructions:
		// the first CD is the head in the flat tree, therefore
		// binarization causes left-branches structures.
		if (node.getFirstChild().isHead()) {
			TreeNode anch = headPath;
			TreeNode curr = anch;
			
			while (curr != node) {
				TreeNode currPa = curr.getParent();
				TreeNode head = currPa.getHeadChild();
				
				if (currPa.getChildren().size() == 1) {
					currPa.setBottom(lex);
					lex.addNode(currPa);
					head.setTop(lex);
					lex.addNode(head);
				}
				
				for (TreeNode child : currPa.getChildren()) {
					if (child.isHead()) continue;
					
					// looking for the foot
					if (child.getFunctionalCategory().matches("CJ") 
							&& lex.getFoot() == null 
							&& !child.getPhrasalCategory().equals("TRUNC")) {
						child.setTop(lex);
						lex.addNode(child);
						lex.setFoot(child);
						
						currPa.setBottom(lex);
						lex.addNode(currPa);
						head.setTop(lex);
						lex.addNode(head);
					}
					// treat conjuncts and punctuation (except sentence-final
					// punctuation) as arguments
					else if (child.getFunctionalCategory().matches("CJ|CD") 
							|| (Utils.isPunct(child.getPhrasalCategory()) 
									&& !Utils.isRightPunctPeriphery(currTigerTree, child))) {
						
						LexEntry lexNew = new LexEntry(child, multiAnchoredTrees);
						if (child.getFunctionalCategory().matches("CD")) {
							child.setNodeStatus(NodeStatus.MAARG);
							lex.addAssociate(lexNew);
						}
						else {	
							child.setNodeStatus(NodeStatus.ARG);
						}
						child.setTop(lex);
						lex.addNode(child);
						extract(child, lexNew);
						
						currPa.setBottom(lex);
						lex.addNode(currPa);
						head.setTop(lex);
						lex.addNode(head);
					}
					else {
						// modifier
						child.setNodeStatus(NodeStatus.MOD);
						TreeNode foot = extractAdj(child, new LexEntry(currPa, multiAnchoredTrees));
					}
				}
				curr = curr.getParent();
			}
			// we arrived at the anchor
			anch.setBottom(lex);
			lex.addNode(anch);
			lex.setLeafLink((LeafNode)anch);
		}
		
		else {
			while (!node.isLeaf()) {
				TreeNode head = node.getHeadChild();
				
				if (node.getChildren().size() == 1) {
					node.setBottom(lex);
					lex.addNode(node);
					head.setTop(lex);
					lex.addNode(head);
					node = head;
				}
				
				for (TreeNode child : node.getChildren()) {
					if (child.isHead()) continue;
					
					// looking for the foot
					if (child.getFunctionalCategory().matches("CJ") 
							&& lex.getFoot() == null 
							&& !child.getPhrasalCategory().equals("TRUNC")) {
						child.setTop(lex);
						lex.addNode(child);
						lex.setFoot(child);
						
						node.setBottom(lex);
						lex.addNode(node);
						head.setTop(lex);
						lex.addNode(head);
						node = head;
					}
					// treat conjuncts and punctuation (except sentence-final
					// punctuation) as arguments
					else if (child.getFunctionalCategory().matches("CJ|CD") 
							|| (Utils.isPunct(child.getPhrasalCategory()) 
									&& !Utils.isRightPunctPeriphery(currTigerTree, child))) {
						child.setNodeStatus(NodeStatus.ARG);
						child.setTop(lex);
						lex.addNode(child);
						extract(child, new LexEntry(child, multiAnchoredTrees));
						
						node.setBottom(lex);
						lex.addNode(node);
						head.setTop(lex);
						lex.addNode(head);
						node = head;
					}
					else {
						// modifier
						child.setNodeStatus(NodeStatus.MOD);
						TreeNode foot = extractAdj(child, new LexEntry(node, multiAnchoredTrees));
						node = foot;
					}
				}
			}
			// we arrived at the anchor
			node.setBottom(lex);
			lex.addNode(node);
			lex.setLeafLink((LeafNode)node);
		}
		
		currLexicon.add(lex);
		return lex.getFoot();
	}

	private boolean footAvailable(TreeNode node) {
		while (!node.isLeaf()) {
			for (TreeNode child : node.getChildren()) {
				if (child.isHead()) continue;
				if (child.getFunctionalCategory().matches("CJ")
						&& !child.getPhrasalCategory().equals("TRUNC")) {
					return true;
				}
			}
			node = node.getHeadChild();
		}
		return false;
	}

	private TreeNode extractAdj(TreeNode child, LexEntry lex) {
		lex.setAdjTree();
		child.setTop(lex);
		lex.addNode(child);
		//root
		TreeNode pa = child.getParent();
		pa.setBottom(lex);
		lex.addNode(pa);
		//foot
		pa.getHeadChild().setTop(lex);
		lex.addNode(pa.getHeadChild());
		pa.getHeadChild().isFoot();
		lex.setFoot(pa.getHeadChild());
		
		LeafNode leaf = extractSubtree(child, lex, true);
		lex.setLeafLink(leaf);
		currLexicon.add(lex);
		return lex.getFoot();
	}
}
